﻿unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, XPMan, Contnrs, shellapi,
  SFmpqapi, StrUtils;

const
  TFTMapFilter = 'Warcraft III Frozen Throne Map|*.w3x';
  ROCMapFilter = 'Warcraft III Reign of Chaos Map|*.w3m';
  BufferSize = 65536;

type
  TWTSEntry = class(TObject)
  private
    FIndex: integer;
  public
    Text: TStringList;
    constructor Create;
    destructor Destroy; override;
    property index: integer read FIndex write FIndex;
  end;
  TMainForm = class(TForm)
    LogMemo: TMemo;
    ControlsPanel: TPanel;
    TargetButton: TButton;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    AboutButton: TButton;
    VerboseCheck: TCheckBox;
    SourceButton: TButton;
    ProcessButton: TButton;
    NoCompactCheck: TCheckBox;
    procedure ProcessButtonClick(Sender: TObject);
    procedure TargetButtonClick(Sender: TObject);
    procedure AboutButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SourceButtonClick(Sender: TObject);
  private
    { Private-Deklarationen }
    LoadFileName: string;
    SaveFileName: string;
    SkinStream: TMemoryStream;
    SkinAvailable: boolean;
    JStream: TMemoryStream;
    JAvailable: boolean;
    JAtScripts: boolean;
    W3AStream: TMemoryStream;
    W3AAvailable: boolean;
    W3HStream: TMemoryStream;
    W3HAvailable: boolean;
    W3IStream: TMemoryStream;
    W3IAvailable: boolean;
    W3QStream: TMemoryStream;
    W3QAvailable: boolean;
    W3TStream: TMemoryStream;
    W3TAvailable: boolean;
    W3UStream: TMemoryStream;
    W3UAvailable: boolean;
    WTSStream: TMemoryStream;
    WTSFile: TStringList;
    WTSData: TObjectList;
    WTSRemapTable: TStringList;
    procedure AddLog(S: string);
    procedure ExtractTStrings;
    procedure WriteWTS;
    procedure AddFiles;
    procedure LoadWTS;
    procedure InlineStrings;
  public
    { Public-Deklarationen }
  end;

var
  MainForm: TMainForm;

implementation

uses AboutUnit;

{$R *.dfm}

{ TWTSEntry }

constructor TWTSEntry.Create;
begin
  inherited;
  Text := TStringList.Create;
end;

destructor TWTSEntry.Destroy;
begin
  Text.Free;
  inherited;
end;

{ TMainForm }

procedure TMainForm.SourceButtonClick(Sender: TObject);
var
  ext: string;
begin
  if OpenDialog.Execute then LoadFileName := OpenDialog.FileName else exit;
  ext := ExtractFileExt(OpenDialog.Filename);
  if ext = '.w3m' then begin
    SaveDialog.Filter := ROCMapFilter;
    SaveDialog.DefaultExt := 'w3m';
  end else if ext = '.w3x' then begin
    SaveDialog.Filter := TFTMapFilter;
    SaveDialog.DefaultExt := 'w3x';
  end else begin
    Application.MessageBox('Only Warcraft III maps can be loaded!','Error',16);
    exit;
  end;
  LogMemo.Clear;
  LogMemo.WordWrap := false;
  AddLog('Selected "'+OpenDialog.FileName+'" as source file');
  SaveDialog.Filename := OpenDialog.Filename;
  TargetButton.Enabled := True;
  ProcessButton.Enabled := False;
end;

procedure TMainForm.AddLog(S: string);
begin
  LogMemo.Lines.Add(S);
  Application.ProcessMessages;
end;

procedure TMainForm.ExtractTStrings;
var
  i: Integer;
  entry: TWTSEntry;
  s: string;
begin
  AddLog('Parsing war3map.wts...');
  WTSData.Clear;
  i := 0;
  while true do
  begin
    while (i < WTSFile.Count) and (Pos('STRING ', WTSFile[i]) = 0) do inc(i);
    if i = WTSFile.Count then break;
    entry := TWTSEntry.Create;
    s := WTSFile[i];
    delete(s, 1, 7);
    entry.index := StrToInt(s);
    repeat
      inc(i);
      s := WTSFile[i];
    until Pos('{', s) = 1;
    repeat
      inc(i);
      s := WTSFile[i];
      if Pos('}', s) = 1 then break;
      entry.Text.Add(s);
    until false;
    WTSData.Add(entry);
    if VerboseCheck.Checked then
              AddLog('          Found string '+IntToStr(entry.index));
  end;
  WTSFile.Clear;
  AddLog('     '+IntToStr(WTSData.Count) + ' strings found');
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  LogMemo.WordWrap := true;
  WTSFile := TStringList.Create;
  WTSData := TObjectList.Create(true);
  WTSRemapTable := TStringList.Create;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  WTSFile.Free;
  WTSData.Free;
  WTSRemapTable.Free;
end;

procedure TMainForm.WriteWTS;
var
  i,j: integer;
  entry: TWTSEntry;
begin
  WTSStream.Clear;
  AddLog('Writing new war3map.wts...');
  if WTSData.Count > 0 then begin
    for i := 0 to WTSData.Count - 1 do begin
      entry := WTSData.Items[i] as TWTSEntry;
      with entry do begin
        if i = 0 then WTSFile.Add('ï»¿STRING '+IntToStr(index))
          else WTSFile.Add('STRING '+IntToStr(index));
        WTSFile.Add('{');
        for j := 0 to Text.Count - 1 do WTSFile.Add(Text[j]);
        WTSFile.Add('}');
        WTSFile.Add('');
        if VerboseCheck.Checked then
                AddLog('          Added string '+IntToStr(index));
      end;
    end;
  WTSFile.SaveToStream(WTSStream);
  WTSFile.Clear;
  AddLog('     Done');
  end else begin
    AddLog('     No strings to write');
  end;
end;

procedure TMainForm.AddFiles;
var
  maphandle: MPQHANDLE;

procedure WriteToMPQ(const hMPQ: MPQHANDLE; filename: string; const AStream: TMemoryStream; const AAvailable: boolean);
begin
  if AAvailable then begin
    if JAtScripts and (filename = 'war3map.j') then begin
      filename := 'Scripts\war3map.j';
      MPQDeleteFile(maphandle,PChar(filename));
    end;
    MPQAddFileFromBufferEx(hMPQ,AStream.Memory,AStream.Size,PChar(filename), MAFA_COMPRESS or MAFA_REPLACE_EXISTING, MAFA_COMPRESS_DEFLATE, Z_BEST_COMPRESSION);
    AddLog('     Added '+filename+' ('+IntToStr(AStream.Size)+' bytes)');
    AStream.Free;
  end;
end;

begin
  AddLog('Writing new map file...');
  if LoadFileName <> SaveFileName then begin
    CopyFile(PChar(LoadFileName), PChar(SaveFileName), False);
    AddLog('     Copied old map file');
  end;
  maphandle := MPQOpenArchiveForUpdate(PChar(SaveFileName),MOAU_OPEN_EXISTING,0);
  try
    if WTSStream.Size = 0 then begin
      MPQDeleteFile(maphandle,'war3map.wts');
      AddLog('     Removed war3map.wts');
    end else WriteToMPQ(maphandle,'war3map.wts',WTSStream,true);
    WriteToMPQ(maphandle,'war3map.j',JStream,JAvailable);
    WriteToMPQ(maphandle,'war3map.w3a',W3AStream,W3AAvailable);
    WriteToMPQ(maphandle,'war3map.w3h',W3HStream,W3HAvailable);
    WriteToMPQ(maphandle,'war3map.w3i',W3IStream,W3IAvailable);
    WriteToMPQ(maphandle,'war3map.w3q',W3QStream,W3QAvailable);
    WriteToMPQ(maphandle,'war3map.w3t',W3TStream,W3TAvailable);
    WriteToMPQ(maphandle,'war3map.w3u',W3UStream,W3UAvailable);
    WriteToMPQ(maphandle,'war3mapSkin.txt',SkinStream,SkinAvailable);
    if MPQDeleteFile(maphandle,'(attributes)') then AddLog('     Removed attributes file');
    if not NoCompactCheck.Checked then begin
      AddLog('Compacting new map file...');
      MPQCompactArchive(maphandle);
      AddLog('     Done');
    end;
  finally
    MPQCloseUpdatedArchive(maphandle,0);
  end;
  AddLog('Everything done.');
end;

procedure TMainForm.LoadWTS;
var
  s: string;
begin
  AddLog('Loading war3map.wts...');
  WTSStream.Position := 0;
  WTSFile.Clear;
  WTSFile.LoadFromStream(WTSStream);
  s := WTSFile.Strings[0];
  Delete(s, 1, 3);
  WTSFile[0] := s;
  AddLog('     '+IntToStr(WTSFile.Count) + ' lines loaded');
end;

procedure TMainForm.AboutButtonClick(Sender: TObject);
begin
  AboutForm.ShowModal;
end;

procedure TMainForm.InlineStrings;
var
  i,oldcount: integer;
  entry: TWTSEntry;
  NotToDelete: TStringList;

procedure InsertIntoStream(const AStream: TMemoryStream; var Buffer; const Count, Offset: Integer ); inline;
var
  stream: TMemoryStream;
begin
  stream := TMemoryStream.Create;
  try
    AStream.Position := Offset;
    stream.Copyfrom(AStream,AStream.Size - Offset);
    AStream.Position := Offset;
    AStream.Write(Buffer,Count);
    AStream.CopyFrom(Stream,0);
  finally
    stream.Free;
  end;
end;

procedure DeleteFromStream(const AStream: TMemoryStream; const Count, Offset: Integer ); inline;
var
  stream: TMemoryStream;
begin
  stream := TMemoryStream.Create;
  try
    AStream.Position := Offset + Count;
    stream.Copyfrom(AStream,AStream.Size - (Offset + Count));
    AStream.Position := Offset;
    AStream.CopyFrom(Stream,0);
    AStream.SetSize(AStream.Size - Count);
  finally
    stream.Free;
  end;
end;


procedure SearchStream(const filename: string; const AAvailable: boolean; const AStream: TMemoryStream);
var
  i,j,k,l,position,offset,changes: integer;
  s: string;
  c: char;
  stream: TStringStream;
  entry: TWTSEntry;
begin
  stream := TStringStream.Create('');
  try
    if AAvailable then begin
      changes := 0;
      offset := 0;
      entry := nil;
      stream.CopyFrom(AStream,0);
      position := Pos('TRIGSTR_',stream.DataString) - 1;
      if position > 0 then begin
        repeat
          AStream.Seek(position + 8 + offset,soFromBeginning);
          AStream.Read(c,1);
          i := 0;
          s := '';
          while c in ['0'..'9'] do begin
            inc(i);
            s := s + c;
            AStream.Read(c,1);
          end;
          j := StrToInt(s);
          for k := 0 to WTSData.Count - 1 do begin
            entry := WTSData.Items[k] as TWTSEntry;
            if entry.index = j then break;
          end;
          if filename='war3map.j' then begin
              s := StringReplace(StringReplace(StringReplace(entry.Text.Text,'\','\\',[rfReplaceAll]),'"','\"',[rfReplaceAll]),#13#10,'\r\n',[rfReplaceAll]);
            if length(s) >= 4 then Delete(s,length(s)-3,4);
          end else begin
            s := entry.Text.Text;
            if length(s) >= 2 then Delete(s,length(s)-1,2);
          end;
          if (length(s) < 1024)
          and ((entry.index <> j)
          or (entry.text.count <= 1)
          or (filename <> 'war3mapSkin.txt')) then begin
            if (entry.index <> j) or (entry.Text.Count = 0) then s := '';
            AStream.Seek(position + offset,soFromBeginning);
            if (i+8) > length(s) then begin
              AStream.Write(PChar(s)^,length(s));
              DeleteFromStream(AStream,(i+8)-length(s),AStream.Position);
            end else begin
              AStream.Write(PChar(s)^,(i+8));
              Delete(s,1,i+8);
              InsertIntoStream(AStream,PChar(s)^,length(s),AStream.Position);
            end;
            AStream.Position := position + offset + length(entry.Text[0]);
            inc(changes);
            if VerboseCheck.Checked then begin
              if entry.index <> j then
                AddLog('          Removed reference to non-existant string '+IntToStr(j)+' in '+filename)
              else if entry.Text.count > 1 then
                AddLog('          Inlined string '+IntToStr(j)+' with '+IntToStr(entry.Text.count)+' lines into '+filename)
              else AddLog('          Inlined string '+IntToStr(j)+' into '+filename);
            end;
          end else NotToDelete.Add(IntToStr(j));
          stream.Size := 0;
          offset := AStream.Position;
          stream.CopyFrom(AStream,AStream.Size - offset);
          position := Pos('TRIGSTR_',stream.DataString) - 1;
        until position = -1;
        AddLog('     '+IntToStr(changes) + ' strings inlined into '+filename);
      end else AddLog('     No strings to inline into '+filename);
    end;
  finally
    stream.free;
  end;
end;

begin
  AddLog('Inlining trigger strings...');
  oldcount := WTSData.Count;
  NotToDelete := TStringList.Create;
  try
    SearchStream('war3map.j',JAvailable,JStream);
    SearchStream('war3map.w3a',W3AAvailable,W3AStream);
    SearchStream('war3map.w3h',W3HAvailable,W3HStream);
    SearchStream('war3map.w3i',W3IAvailable,W3IStream);
    SearchStream('war3map.w3q',W3QAvailable,W3QStream);
    SearchStream('war3map.w3t',W3TAvailable,W3TStream);
    SearchStream('war3map.w3u',W3UAvailable,W3UStream);
    SearchStream('war3mapSkin.txt',SkinAvailable,SkinStream);
    AddLog('Removing inlined strings from war3map.wts...');
    for i := WTSData.Count - 1 downto 0 do begin
      entry := WTSData.Items[i] as TWTSEntry;
      if NotToDelete.IndexOf(IntToStr(entry.index)) = -1 then begin
        WTSData.Delete(i);
        if VerboseCheck.Checked then
          AddLog('          Removed string '+IntToStr(i));
      end;
    end;
    AddLog('     Removed '+IntToStr(oldcount - WTSData.Count)+' strings ('+IntToStr(WTSData.Count)+' strings retained)');
  finally
    NotToDelete.Free;
  end;
end;

procedure TMainForm.TargetButtonClick(Sender: TObject);
begin
  if SaveDialog.Execute then SaveFileName := SaveDialog.FileName else exit;
  AddLog('Selected "'+SaveDialog.FileName+'" as target file');
  ProcessButton.Enabled := True;
end;

procedure TMainForm.ProcessButtonClick(Sender: TObject);
var
  maphandle: MPQHANDLE;
  filehandle: MPQHANDLE;
  buffer: pointer;
  bytesread: DWORD;
begin
  SourceButton.Enabled := False;
  TargetButton.Enabled := False;
  ProcessButton.Enabled := False;
  VerboseCheck.Enabled := False;
  NoCompactCheck.Enabled := False;
  AddLog('Extracting files from map...');
  if not SFileOpenArchive(PChar(LoadFileName),0,0,@maphandle) then begin
    raise Exception.Create('Error opening map file!');
  end;
  GetMem(buffer,buffersize);
  try
    if not SFileOpenFileEx(maphandle,'war3map.wts',0,@filehandle) then begin
      raise Exception.Create('war3map.wts not found inside map!');
      SFileCloseArchive(maphandle);
    end;
    WTSStream := TMemoryStream.Create;
    repeat
      SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
      WTSStream.Write(buffer^,bytesread);
    until bytesread < buffersize;
    SFileCloseFile(filehandle);
    AddLog('     Extracted war3map.wts ('+IntToStr(WTSStream.Size)+' bytes)');
    if SFileOpenFileEx(maphandle,'war3map.j',0,@filehandle) then begin
      JAvailable := True;
      JAtScripts := False;
      JStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        JStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3map.j ('+IntToStr(JStream.Size)+' bytes)');
    end else if SFileOpenFileEx(maphandle,'Scripts\war3map.j',0,@filehandle) then begin
      JAvailable := True;
      JAtScripts := True;
      JStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        JStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted Scripts\war3map.j ('+IntToStr(JStream.Size)+' bytes)');
    end else begin
      JAvailable := False;
      Addlog('     war3map.j not found inside map');
    end;
    if SFileOpenFileEx(maphandle,'war3map.w3a',0,@filehandle) then begin
      W3AAvailable := True;
      W3AStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        W3AStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3map.w3a ('+IntToStr(W3AStream.Size)+' bytes)');
    end else begin
      W3AAvailable := False;
      Addlog('     war3map.w3a not found inside map');
    end;
    if SFileOpenFileEx(maphandle,'war3map.w3h',0,@filehandle) then begin
      W3HAvailable := True;
      W3HStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        W3HStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3map.w3h ('+IntToStr(W3HStream.Size)+' bytes)');
    end else begin
      W3HAvailable := False;
      Addlog('     war3map.w3h not found inside map');
    end;
    if SFileOpenFileEx(maphandle,'war3map.w3i',0,@filehandle) then begin
      W3IAvailable := True;
      W3IStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        W3IStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3map.w3i ('+IntToStr(W3IStream.Size)+' bytes)');
    end else begin
      W3IAvailable := False;
      Addlog('     war3map.w3i not found inside map');
    end;
    if SFileOpenFileEx(maphandle,'war3map.w3q',0,@filehandle) then begin
      W3QAvailable := True;
      W3QStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        W3QStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3map.w3q ('+IntToStr(W3QStream.Size)+' bytes)');
    end else begin
      W3QAvailable := False;
      Addlog('     war3map.w3q not found inside map');
    end;
    if SFileOpenFileEx(maphandle,'war3map.w3t',0,@filehandle) then begin
      W3TAvailable := True;
      W3TStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        W3TStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3map.w3t ('+IntToStr(W3TStream.Size)+' bytes)');
    end else begin
      W3TAvailable := False;
      Addlog('     war3map.w3t not found inside map');
    end;
    if SFileOpenFileEx(maphandle,'war3map.w3u',0,@filehandle) then begin
      W3UAvailable := True;
      W3UStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        W3UStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3map.w3u ('+IntToStr(W3UStream.Size)+' bytes)');
    end else begin
      W3UAvailable := False;
      Addlog('     war3map.w3u not found inside map');
    end;
    if SFileOpenFileEx(maphandle,'war3mapSkin.txt',0,@filehandle) then begin
      SkinAvailable := True;
      SkinStream := TMemoryStream.Create;
      repeat
        SFileReadFile(filehandle,buffer,buffersize,@bytesread,nil);
        SkinStream.Write(buffer^,bytesread);
      until bytesread < buffersize;
      SFileCloseFile(filehandle);
      AddLog('     Extracted war3mapSkin.txt ('+IntToStr(SkinStream.Size)+' bytes)');
    end else begin
      SkinAvailable := False;
      Addlog('     war3mapSkin.txt not found inside map');
    end;
  finally
    FreeMem(buffer);
    SFileCloseArchive(maphandle);
  end;
  LoadWTS;
  ExtractTStrings;
  if WTSData.Count = 0 then begin
    AddLog('No strings to inline');
    exit;
  end;
  InlineStrings;
  WriteWTS;
  AddFiles;
  SourceButton.Enabled := True;
  TargetButton.Enabled := True;
  ProcessButton.Enabled := True;
  VerboseCheck.Enabled := True;
  NoCompactCheck.Enabled := True;
end;

end.
